// draw-tree.cc
//
// (c) Mark Johnson, 26th June 2004, 23rd October 2004, 20th March 2008, 
//     2nd July 2009, 22nd July 2010, 11th January 2012, 19th April 2012
//
// A program that uses the generic tree-drawing code in draw-tree.h
// and the tree package in tree.h to draw trees.
//
// It uses the Cairo library to create PDFs.

#include "custom_allocator.h"  // must be included first

const char version[] = "draw-tree version of 19th April, 2012";

const char usage[] =
"\nUsage:\n"
"\n"
"draw-tree [-s <size>] [-c] [-d] [-e] [-h] [-p] [-P] [-t] [-w] [-W] [-x]\n"
"\n"
"reads in trees from stdin in Lisp/Treebank format and produces graphic images of them.\n"
"\n"
"The input should contain one or more [<filestem>] <tree> pairs, where:\n"
"\n"
"  <filestem> is an optional string not beginning with '(' that name this tree's output, and\n"
"  <tree> is a tree in Lisp/Treebank format.\n"
"\n"
"Whitespace (including blank lines) is ignored in the input.\n"
"\n"
"The output that draw-tree produces depends on the command-line flags:\n"
"\n"
"  -c displays node labels beginning and ending with '!' in red color\n"
"     and beginning and ending with '#' in blue color\n"
"\n"
"  -d writes each tree to stdout, one per line (useful for debugging)\n"
"\n"
"  -e produces an encapsulated postscript file with the name \"<filestem>.eps\" for each <tree>\n"
"\n"
"  -f produces a 'flat' tree, where all terminals are on the same line\n"
"\n"
"  -h prints using Helvetica rather than Times if possible\n"
"\n"
"  -p pretty-prints each tree to stdout, over several lines\n"
"\n"
"  -P produces a PDF file with the name \"<filestem>.pdf\" for each <tree>\n"
"\n"
"  -t produces a TikZ file with the name \"<filestem>.tex\" for each <tree>\n"
"\n"
"  -x produces an Xfig file with the name \"<filestem>.fig\" for each <tree>\n"
"\n"
"  -w draws each <tree> in its own wish window with title \"<filestem>\"\n"
"\n"
"  -W draws each <tree> in the same wish window, erasing the last tree (this is handy if\n"
"     the input stream is incrementally producing trees that you wish to view)\n"
"\n"
"The -s option sets the point size in the trees that are drawn.\n"
"\n";

#include "draw-tree.h"
#include "tree.h"

#ifdef USECAIRO
#include <cairo/cairo.h>
#include <cairo/cairo-pdf.h>
#endif

#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <string>

static const symbol empty_symbol("");

//! char_width[c] is the width of character c in 100'ths of a pixel
//
const int char_width[] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
  288, 360, 504, 576, 576, 1008, 936, 360, 360, 360, 576, 648, 288, 360, 288, 360, 576, 576, 576, 
  576, 576, 576, 576, 576, 576, 576, 288, 288, 648, 648, 648, 504, 1080, 864, 792, 792, 864, 720, 
  648, 864, 864, 360, 432, 864, 720, 1080, 864, 864, 648, 864, 792, 648, 720, 864, 864, 1152, 864, 
  864, 720, 360, 360, 360, 576, 576, 360, 504, 576, 504, 576, 504, 432, 576, 576, 360, 288, 576, 
  360, 936, 576, 576, 576, 576, 432, 432, 360, 576, 576, 864, 576, 576, 504, 576, 216, 576, 648, 300, 300
};

//! interface{} is the interface between the tree{} nodes and 
//! draw_tree{}
//
template <typename node_type_=tree>
struct tree_interface {
  typedef node_type_ node_type;
  
  //! nodes_type{} is a Sequence (in the STL sense) of nodes.
  //
  struct nodes_type {

    struct const_iterator {
      const_iterator(const node_type* np) : np(np) { }
      const node_type* np;
      const node_type& operator*() const { return *np; }
      const node_type* operator->() const { return np; }
      const_iterator operator++() { return const_iterator(np = np->next); }
      bool operator!=(const const_iterator& other) const { return np != other.np; }
      bool operator==(const const_iterator& other) const { return np == other.np; }
    }; // tree_interface::nodes_type::const_iterator{}

    const node_type* node;
    nodes_type(const node_type* node) : node(node) { }
    const_iterator begin() const { return const_iterator(node); }
    const_iterator end() const { return NULL; }
    bool empty() const { return node == NULL; }
  }; // tree_interface::nodes_type{}

  nodes_type children(const node_type& parent) const { return nodes_type(parent.child); }

  const char* label_color(const node_type& n) const {
    static const char red[] = "red";
    static const char blue[] = "blue";
    if (colorize) {
      std::ostringstream os;
      os << n.label;
      std::string str = os.str();
      if (str.size() > 2 && str[0] == str[str.size()-1]) {
	if (str[0] == '!')
	  return red;
	else if (str[0] == '#')
	  return blue;
      }
    }
    return NULL;
  }  //tree_interface::label_color()

  std::string label_string(const node_type& n) const {
    std::ostringstream os;
    os << n.label;
    std::string str = os.str();
    if (label_color(n)) 
      return str.substr(1, str.size()-2);
    else
      return str;
  } // tree_interface::label_string()

  // The stuff below gives approximate widths and heights in points for a 10 pt font
  //
  int fontsize;
  bool colorize;
  bool helvetica;

  tree_interface(int fontsize=10, bool colorize=false, bool helvetica=false) 
    : fontsize(fontsize), colorize(colorize), helvetica(helvetica) { }

  int label_width(const node_type& n) const {
    const std::string s = label_string(n);
    int w = 0;
    for (std::string::const_iterator it = s.begin(); it != s.end(); ++it)
      if ((size_t) *it < sizeof(char_width)/sizeof(char_width[0]))
	w += char_width[int(*it)];
    return fontsize*(w+50)/1200;
  } // tree_interface::label_width()

  int label_height(const node_type& n) const { return fontsize; }
  int label_descent(const node_type& n) const { return 2; }
  int gap_width() const { return fontsize/2; }
  int gap_height() const { return fontsize/2; }
  int border_width() const { return fontsize/2; }
  int border_height() const { return fontsize/2; }
};  // tree_interface{}


//! An eps_tree_interface{} saves trees as eps files.
//
template <typename node_type=tree>
struct eps_tree_interface : public tree_interface<node_type> {

  std::ofstream os;
  int ymax;

  int border_width() const { return 2; }
  int border_height() const { return 1; }

  //! eps_tree_interface() should be initialized with the name of the file
  //! to which the tree should be saved.
  //
  eps_tree_interface(const char* filename, int fontsize=10, bool colorize=false, bool helvetica=false) 
    : tree_interface<node_type>(fontsize, colorize, helvetica), os(filename), ymax(0) { 
    os << "%!PS-Adobe-2.0 EPSF-2.0\n"
          "%%Title: " << filename << "\n"
          "%%Creator: draw-tree\n";
  }  // eps_tree_interface::eps_tree_interface()

  void drawing_area(int width, int height) {
    ymax = height;
    os << "%%BoundingBox: 0 0 " << width << ' ' << height << '\n'
       << "%%Magnification: 1.0000\n"
          "%%EndComments\n\n";
    if (this->helvetica)
      os << "/Helvetica findfont\n";
    else
      os << "/Times-Roman findfont\n";
    os << this->fontsize << " scalefont\nsetfont\n\n";
    os << "0.5 setlinewidth\n\n";
    os << "/cshow % stack: string\n{ dup stringwidth\n  pop\n  -2 div\n  0 rmoveto\n  show\n} def\n\n";
  }  // eps_tree_interface::drawing_area()

  void drawline(int x0, int y0, int x1, int y1) {
    os << "newpath " << x0 << ' ' << ymax-y0 << " moveto " << x1 << ' ' << ymax-y1 << " lineto stroke\n";
  } // eps_tree_interface::drawline()

  void drawlabel(const node_type& n, int x, int y) {
    std::string s = label_string(n);
    os << x << ' ' << ymax-label_height(n)-y << " moveto ("; 
    for (std::string::const_iterator it = s.begin(); it != s.end(); ++it)
      if (isprint(*it)) {
	if (*it == '(' || *it == ')' || *it == '\\')
	  os << '\\' << *it;
	else
	  os << *it;
      }
      else 
	os << '\\' << std::setbase(8) << int(*it) << std::setbase(10);
    os << ")";
    const char* color = label_color(n);
    if (color && strcmp(color, "red") == 0)
      os << " 1.000 0.000 0.000 setrgbcolor cshow 0 setgray\n";
    else if (color && strcmp(color, "blue") == 0)
      os << " 0.000 0.000 1.000 setrgbcolor cshow 0 setgray\n";
    else
      os << " cshow\n";
  } // eps_tree_interface::drawlabel()

  void flush() { os << "\nshowpage" << std::endl; }

};  // eps_tree_interface{}


//! An xfig_tree_interface{} saves trees as xfig files.
//
template <typename node_type=tree>
struct xfig_tree_interface : public tree_interface<node_type> {
  enum {x_scale = 20, y_scale = 15};

  std::ofstream os;

  //! xfig_tree_interface() should be initialized with the name of the file
  //! to which the tree should be saved.
  //
  xfig_tree_interface(const char* filename, int fontsize=10, bool colorize=false, bool helvetica=false) 
    : tree_interface<node_type>(fontsize, colorize, helvetica), os(filename) { 
    os << "#FIG 3.2\n# " << filename 
       << ": Xfig commands produced by draw-tree.cc, written by Mark Johnson" << std::endl;
    os << "Portrait\nCenter\nInches\nLetter\n100.00\nSingle\n-2\n1200 0\n";
  }  // xfig_tree_interface::xfig_tree_interface()

  void drawing_area(int width, int height) {
    os << "# width = " << width*x_scale << ", height = " << height*y_scale << std::endl;
  }  // xfig_tree_interface::drawing_area()

  void drawline(int x0, int y0, int x1, int y1) {
    os << "2 1 0 1 0 7 50 0 -1 0.000 0 0 -1 0 0 2\n" 
       << "\t " << x0*x_scale << ' ' << y0*y_scale
       << ' ' << x1*x_scale << ' ' << y1*y_scale << std::endl;
  } // xfig_tree_interface::drawline()

  void drawlabel(const node_type& n, int x, int y) {
    int color = 0;
    const char* colorstr = label_color(n);
    if (colorstr && strcmp(colorstr, "red") == 0)
      color = 4;
    else if (colorstr && strcmp(colorstr, "blue") == 0)
      color = 1;
    os << "4 1 " 
       << color 
       << " 50 0 "
       << (this->helvetica ? "4 " : "0 ")
       << this->fontsize << " 0 2 135 "
       << label_width(n)*x_scale << ' ' << x*x_scale 
       << ' ' << (y+label_height(n))*y_scale 
       << ' ' << label_string(n) << "\\001" << std::endl;
  } // xfig_tree_interface::drawlabel()

  void flush() const { }

};  // xfig_tree_interface{}


//! The wish_tree_interface{} constructs a wish window and displays
//! trees in it.
//
template <typename node_type=tree>
struct wish_tree_interface : public tree_interface<node_type> {

  FILE* fp;
  std::string title;

  wish_tree_interface(const char* title="", int fontsize=10, bool colorize=false, bool helvetica=false) 
    : tree_interface<node_type>(fontsize, colorize, helvetica), fp(NULL), title(title) { }

  void drawing_area(int width, int height) {
    if (fp == NULL) {
      fp = popen("wish", "w");
      fprintf(fp, "wm minsize . 20 20\n");
      fprintf(fp, "frame .f\n");
      fprintf(fp, "canvas .c -xscrollcommand {.hscroll set} -yscrollcommand {.vscroll set} -scrollregion {-10p -10p 1000p 1000p} -bg white\n");
      fprintf(fp, "scrollbar .hscroll -orient horizontal\n");
      fprintf(fp, "pack .hscroll -side bottom -fill x -in .f\n");
      fprintf(fp, "pack .c -side top -expand 1 -fill both -in .f\n");
      fprintf(fp, "scrollbar .vscroll -orient vertical\n");
      fprintf(fp, "pack .vscroll -side right -fill y\n");
      fprintf(fp, "pack .f -side left -expand 1 -fill both\n");
      fprintf(fp, ".vscroll config -command {.c yview}\n");
      fprintf(fp, ".hscroll config -command {.c xview}\n");
    }
    else 
      fprintf(fp, ".c delete all\n");
    if (title.empty())
      fprintf(fp, "wm title . \"Draw-tree wish display\"\n");
    else
      fprintf(fp, "wm title . \"%s\"\n", title.c_str());
    fprintf(fp, ".c config -scrollregion {-10p -10p %dp %dp}\n", width+10, height+10);
  }  // wish_tree_interface::drawing_area()

  void drawline(int x0, int y0, int x1, int y1) {
    fprintf(fp, ".c create line %dp %dp %dp %dp\n", x0, y0, x1, y1);
  } // wish_tree_interface::drawline()

  void drawlabel(const node_type& n, int x, int y) {
    static const char font0[] = "-*-times-medium-r-normal--*-";
    static const char hfont0[] = "-*-helvetica-medium-r-normal--*-";
    static const char font1[] = "0-*";
    const char* color = label_color(n);
    fprintf(fp, ".c create text %dp %dp -text {%s} -anchor n -font \"%s%d%s\"",
	    x, y+1, label_string(n).c_str(), (this->helvetica ? hfont0 : font0), this->fontsize, font1);
    if (color) 
      fprintf(fp, " -fill %s\n", color);
    else
      fprintf(fp, "\n");
  } // wish_tree_interface::drawlabel()

  void flush() const { fflush(fp); }

};  // wish_tree_interface{}

#ifdef USECAIRO

//! An pdf_tree_interface{} saves trees as pdf files.
//
template <typename node_type=tree>
struct pdf_tree_interface : public tree_interface<node_type> {

  int ymax;
  cairo_surface_t* csurface;
  cairo_t* cr;

  int border_width() const { return 2; }
  int border_height() const { return 1; }

  //! pdf_tree_interface() should be initialized with the name of the file
  //! to which the tree should be saved.
  //
  pdf_tree_interface(const char* filename, int fontsize=10, bool colorize=false, bool helvetica=false) 
    : tree_interface<node_type>(fontsize, colorize, helvetica), 
      ymax(0), csurface(NULL), cr(NULL) 
  { 
    assert(csurface == NULL);
    csurface = cairo_pdf_surface_create(filename, 0, 0);
    assert(cairo_surface_status(csurface) == CAIRO_STATUS_SUCCESS);
    cr = cairo_create(csurface);
    cairo_select_font_face(cr, helvetica ? "Helvetica" : "Times", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
    cairo_set_font_size(cr, fontsize);
    cairo_set_line_width(cr, 0.5);
  }  // pdf_tree_interface::pdf_tree_interface()

  ~pdf_tree_interface() {
    cairo_destroy(cr);
    cairo_surface_destroy(csurface);
  }  // pdf_tree_interface::~pdf_tree_interface()

  int gap_width() const { return this->fontsize/2; }

  int label_width(const node_type& n) const {
    std::string s = label_string(n);
    cairo_text_extents_t te;
    cairo_text_extents(cr, s.c_str(), &te);
    int width = te.width + te.x_bearing + 0.9;
    return width;
  } // pdf_tree_interface::label_width()

  void drawing_area(int width, int height) {
    ymax = height;
    cairo_pdf_surface_set_size(csurface, width, height);
  }  // pdf_tree_interface::drawing_area()

  void drawline(int x0, int y0, int x1, int y1) {
    cairo_move_to(cr, x0, y0);
    cairo_line_to(cr, x1, y1);
    cairo_stroke(cr);
  } // pdf_tree_interface::drawline()

  void drawlabel(const node_type& n, int x, int y) {
    std::string s = label_string(n);
    cairo_font_extents_t fe;
    cairo_font_extents (cr, &fe);
    cairo_text_extents_t te;
    cairo_text_extents(cr, s.c_str(), &te);
    cairo_move_to(cr, x - te.width / 2 - te.x_bearing,
		  y + fe.ascent); //  + te.height + te.y_bearing
    const char* color = label_color(n);
    if (color && strcmp(color, "red") == 0) {
      cairo_set_source_rgb(cr, 1, 0, 0);
      cairo_show_text(cr, s.c_str());
      cairo_set_source_rgb(cr, 0, 0, 0);
    }
    else if (color && strcmp(color, "blue") == 0) {
      cairo_set_source_rgb(cr, 0, 0, 1);
      cairo_show_text(cr, s.c_str());
      cairo_set_source_rgb(cr, 0, 0, 0);
    }
    else
      cairo_show_text(cr, s.c_str());
  } // pdf_tree_interface::drawlabel()

  void flush() const { }

};  // pdf_tree_interface{}
#endif

//! An tikz_tree_interface{} saves trees as .tex files with TikZ commands that draw the tree.
//
template <typename node_type=tree>
struct tikz_tree_interface : public tree_interface<node_type> {

  std::ostream& os;
  int ymax;

  int border_width() const { return 2; }
  int border_height() const { return 1; }

  //! tikz_tree_interface() should be initialized with the name of the file
  //! to which the tree should be saved.
  //
  tikz_tree_interface(std::ostream& os, int fontsize=10, bool colorize=false, bool helvetica=false) 
    : tree_interface<node_type>(fontsize, colorize, helvetica), os(os), ymax(0) { 
    os << "\\tikzstyle{treeline}=[semithick]\n";
    os << "\\tikzstyle{treenonterminal}=[text height=0.5em,text depth=0.1em,anchor=base]\n";
    os << "\\tikzstyle{treeterminal}=[treenonterminal,font=\\itshape]\n";
    os << "\\begin{tikzpicture}[baseline,x=0.1em,y=0.1em]\n";
  }  // tikz_tree_interface::tree_interface()

  void drawing_area(int width, int height) {
    ymax = height;
  }  // tikz_tree_interface::drawing_area()

  void drawline(int x0, int y0, int x1, int y1) {
    os << "\\draw[treeline] (" << x0 << "," << -y0 << ") -- (" << x1 << "," << -y1 << ");\n";
  } // tikz_tree_interface::drawline()

  void drawlabel(const node_type& n, int x, int y) {
    std::string s = label_string(n);
    const char* color = label_color(n);
    os << "\\node[";
    if (n.child == NULL)
      os << "treeterminal";
    else
      os << "treenonterminal";
    if (color)
      os << ",color=" << color;
    os << "] at (" << x << "," << -(y+label_height(n)) << ") {" << s << "};\n";
  } // tikz_tree_interface::drawlabel()

  void flush() { os << "\\end{tikzpicture}\n" << std::endl; }

};  // tikz_tree_interface{}

/*
// tikz_tree() produces a TikZ tree structure.
//
// However, usually nodes in this tree will overlap, so it isn't used any more
//
template <typename label_type>
void tikz_tree(std::ostream& os, const tree_node<label_type>* t, int indent=0)
{
  if (indent == 0) 
    os << "\\begin{tikzpicture}\n \\";

  os << "node {" << t->label << "}";

  if (t->child) {
    int indent2 = indent+2;
    for(tree_node<label_type>* p = t->child; p; p = p->next) {
      os << '\n';
      for (int i = 0; i < indent2; ++i)
	os << ' ';
      os << "child{ ";
      tikz_tree(os, p, indent2);
      os << "}";
    }
  }

  if (indent == 0)
    os << ";\n\\end{tikzpicture}\n";
}
*/

int main(int argc, char** argv) {

  int fontsize = 10;
  float scale = 1.0;
  bool optset = false;
  bool colorize = false;
  bool copy = false;
  bool flat = false;
  bool eps = false;
  bool pdf = false;
  bool helvetica = false;
  bool pretty = false;
  bool tikz = false;
  bool xfig = false;
  bool wish = false;
  bool wish1 = false;

  int c;
  while ((c = getopt(argc, argv, "s:S:cdefhpPtwWx")) != -1)
    switch (c) {
    case 'c':
      colorize = true;
      break;
    case 'd':
      optset = copy = true;
      break;
    case 'e':
      optset = eps = true;
      break;
    case 'f':
      flat = true;
      break;
    case 'h':
      helvetica = true;
      break;
    case 'p':
      optset = pretty = true;
      break;
    case 'P':
      optset = pdf = true;
      break;
    case 's':
      fontsize = atoi(optarg);
      break;
    case 'S':
      scale = atof(optarg);
      break;
    case 't':
      optset = tikz = true;
      break;
    case 'w':
      optset = wish = true;
      break;
    case 'W':
      optset = wish1 = true;
      break;
    case 'x':
      optset = xfig = true;
      break;
    default:
      std::cerr << "# " << version << std::endl;
      std::cerr << "## Error: can't interpret command line argument: -" << char(c) << std::endl;
      std::cerr << usage << std::endl;
      exit(EXIT_FAILURE);
      break;
    }
  
  if (argc != optind) {
    std::cerr << "# " << version << std::endl;
    std::cerr << "## Error: unknown command line argument: " << argv[optind] << std::endl;
    std::cerr << usage << std::endl;
    exit(EXIT_FAILURE);
  }

  if (!optset) { 
    std::cerr << "# Error: no output options set in " << version << "\n" << usage << std::endl;
    exit(EXIT_FAILURE);
  }	
	
  wish_tree_interface<> wi1("", fontsize, colorize, helvetica);  //!< -W wish1 output

  bool named;
  std::string filename;
  int treeno = 0;
  char chr;

  while (std::cin >> chr) {
    std::cin.unget();
    ++treeno;

    if (chr == '(') {
      named = false;
      filename = "tree-";
      std::ostringstream os;
      os << treeno;
      filename += os.str();
    }
    else {
      named = true;
      std::cin >> filename;
    }

    tree* tp;
    std::cin >> tp;

    if (!tp) {
      std::cerr << "Error: failed to read a tree following filename " << filename <<std::endl;
      break;
    }

    if (tp->is_root() || tp->label.cat == empty_symbol) {
      tree* old_tp = tp;
      tp = tp->child;
      old_tp->delete_this();
    }
    
    if (copy)
      std::cout << filename << " " << tp << std::endl;

    if (pretty) {
      display_tree(std::cout, tp);
      std::cout << std::endl;
    }

    if (eps) {
      std::string epsfilename = filename + ".eps";
      eps_tree_interface<> ei(epsfilename.c_str(), fontsize, colorize, helvetica);
      draw_tree<eps_tree_interface<> >::draw(ei, *tp, scale, flat);
    }

    if (pdf) {
#ifdef USECAIRO
      std::string pdffilename = filename + ".pdf";
      pdf_tree_interface<> pi(pdffilename.c_str(), fontsize, colorize, helvetica);
      draw_tree<pdf_tree_interface<> >::draw(pi, *tp, scale, flat);
#endif
    }

    if (tikz) {
      if (named) {
	std::string tikzfilename = filename + ".tex";
	std::ofstream os(tikzfilename.c_str());
	tikz_tree_interface<> ti(os, fontsize, colorize, helvetica);
	draw_tree<tikz_tree_interface<> >::draw(ti, *tp, scale, flat);
      }
      else {
	tikz_tree_interface<> ti(std::cout, fontsize, colorize, helvetica);
	draw_tree<tikz_tree_interface<> >::draw(ti, *tp, scale, flat);
      }
    }

    if (xfig) {
      std::string xfigfilename = filename + ".fig";
      xfig_tree_interface<> xi(xfigfilename.c_str(), fontsize, colorize, helvetica);
      draw_tree<xfig_tree_interface<> >::draw(xi, *tp, scale, flat);
    }

    if (wish) {
      wish_tree_interface<> wi(filename.c_str(), fontsize, colorize, helvetica);
      draw_tree<wish_tree_interface<> >::draw(wi, *tp, scale, flat);
    }

    if (wish1) {
      wi1.title = filename;
      draw_tree<wish_tree_interface<> >::draw(wi1, *tp, scale, flat);
    }

    delete tp;
  }

} // main()
