// draw-tree.h
//
// (c) Mark Johnson, 26th June 2004, 11th January 2012
//
//! draw-tree.h provides a generic tree-drawing package.  It is
//! generic in two senses: first, it is device independent, and
//! second, it works with arbitrary tree data structures.
//!
//! In order to port the tree drawing routine to a new device or
//! tree data structure you should define a new "tree_interface"
//! class.  A tree_interface class needs to define the following:
//!
//! node_type -- the type of nodes
//!
//! nodes_type -- the type of a Sequence (in the STL sense) of nodes
//!
//! A tree_interface object t should have the following methods:
//!
//! const nodes_type& t.children(const node_type& n) -- returns a 
//!      Sequence of the children of n
//!
//! int t.label_width(const node_type& n) -- the width of node n
//!
//! int t.label_height(const node_type& n) -- the height of node n
//!
//! int t.label_descent(const node_type& n) -- space below node n
//!
//! int t.gap_width() -- the minimum gap between nodes
//!
//! int t.gap_height() -- the minimum gap between nodes
//!
//! int t.border_width() -- the width of the border around the tree
//!
//! int t.border_height() -- the height of the border around the tree
//!
//! void t.drawing_area(int width, int height) --
//!      sets the size of the canvas to (width,height).
//!
//! void t.drawlabel(const node_type& n, int x, int y) 
//!      -- draw n at (x,y)
//!
//! void t.drawline(int x0, int y0, int x1, int y1)
//!      -- draw a line from (x0,y0) to (x1,y1)
//!
//! void t.flush() -- complete any buffered drawing commands


#ifndef DRAW_TREE_H
#define DRAW_TREE_H

#include <algorithm>
#include <deque>
#include <utility>
#include <vector>

#include "util.h"

template <typename tree_interface>
class draw_tree {

private:

  // A lot of boring typedefs. 

  typedef typename tree_interface::node_type node_type;
  typedef typename tree_interface::nodes_type nodes_type;

  struct dtnode_type {
    typedef std::vector<dtnode_type> dtnodes_type;
    dtnodes_type children;
    const node_type* nodep;  //!< tree node 
    int dx;                  //!< position relative to parent
    int dy;
    int dcx;		     //!< position of children relative to this node
    dtnode_type(const node_type* nodep = NULL, int dx = 0, int dy = 0, int dcx = 0) 
      : nodep(nodep), dx(dx), dy(dy), dcx(dcx) { }
  }; // draw_tree::dtnode_type{}

  typedef typename dtnode_type::dtnodes_type dtnodes_type;
  typedef std::pair<int,int> II;
  typedef std::deque<II> IIs;

  //! translate() shifts a vector of (x,y) positions by (dx,dy)
  //
  static void translate(int dx, int dy, IIs& xys) {
    for (IIs::iterator it = xys.begin(); it != xys.end(); ++it) {
      it->first += dx;
      it->second += dy;
    }
  }  // draw_tree::translate()

  //! max_overlap() returns the maximum value of Xright-Xleft for the same Y value
  //
  static int max_overlap(const IIs& xys0, const IIs& xys1, bool flat) {
    int max_overlap0 = 0;
    IIs::const_iterator it0 = xys0.begin();
    IIs::const_iterator it1 = xys1.begin();
    int x0 = 0, x1 = 0;
    while (it0 != xys0.end() && it1 != xys1.end()) {
      x0 = it0->first;
      x1 = it1->first;
      max_overlap0 = std::max(max_overlap0, x0 - x1);
      if (it0->second == it1->second) {
	++it0;
	++it1;
      }
      else if (it0->second < it1->second)
	++it0;
      else 
	++it1;
    }
    if (flat) {
      while (it0 != xys0.end())
	max_overlap0 = std::max(max_overlap0, it0++->first - x1);
      while (it1 != xys1.end())
	max_overlap0 = std::max(max_overlap0, x0 - it1++->first);
    }
    return max_overlap0;
  }  // draw_tree::max_overlap()

  //! merge_sides() copys the (x,y) values in xys0 into xys where
  //! y is greater than the last y value in xys.
  //
  static void merge_sides(const IIs& xys0, IIs& xys) {
    if (xys.empty())
      xys = xys0;
    else {
      int last_y = xys.back().second;
      IIs::const_iterator it0 = xys0.begin();
      while (it0 != xys0.end() && it0->second <= last_y)
	++it0;
      xys.insert(xys.end(), it0, xys0.end());
    }
  }  // draw_tree::merge_sides()

  //! merge_sides_flat() copys the (x,y) values in xys0 into xys where
  //! y is greater than the last y value in xys, resetting x as required.
  //
  static void merge_sides_flat(const IIs& xys0, IIs& xys) {
    if (xys.empty()) 
      xys = xys0;
    else if (!xys0.empty())
      xys.back().second = std::max(xys.back().second, xys0.back().second);
  }  // draw_tree::merge_sides()

  //! layout() sets the locations of the dtnodes so they don't
  //! overlap.
  //
  static void layout(tree_interface& ti, const node_type& node,
		     dtnode_type& dtparent, IIs& lefts, IIs& rights,
		     float scale, bool flat) {
    dtparent.children.push_back(dtnode_type(&node));
    dtnode_type& dtnode = dtparent.children.back();
    lefts.clear();
    rights.clear();
    int lx = ti.label_width(node);
    int ly = ti.label_height(node)+ti.label_descent(node);
    int gx = ti.gap_width();
    int gy = (int) (scale*ti.gap_height());
    const nodes_type& children = ti.children(node);
    if (children.empty()) {
      lefts.push_back(II(-lx/2,ly));
      rights.push_back(II(lx/2,ly));
    }
    else {  // !children.empty()
      typename nodes_type::const_iterator cit = children.begin();
      typename nodes_type::const_iterator next_cit = cit; ++next_cit;
      if (next_cit == children.end()) {  // one child
	layout(ti, *cit, dtnode, lefts, rights, scale, flat);
	dtnode_type& dtchild = dtnode.children.back();
	dtchild.dy = ly+gy;
	translate(0, ly+gy, lefts);
	translate(0, ly+gy, rights);
	lefts.push_front(II(-lx/2,ly));
	rights.push_front(II(lx/2,ly));
      }
      else {  // two or more children
	IIs rights0, lefts1, rights1;
	layout(ti, *cit, dtnode, lefts, rights0, scale, flat); // first child
	dtnode_type& dtleftchild = dtnode.children.back();
	dtleftchild.dy = ly+gy;
	while (++next_cit != children.end()) {      
	  ++cit;
	  layout(ti, *cit, dtnode, lefts1, rights1, scale, flat); // second to second to last child
	  dtnode_type& dtchild = dtnode.children.back();
	  int dx = max_overlap(rights0, lefts1, flat);
	  dtchild.dx = dx+gx;
	  dtchild.dy = ly+gy;
	  translate(dx+gx, 0, lefts1);
	  translate(dx+gx, 0, rights1);
	  if (flat) {
	    merge_sides_flat(lefts1, lefts);
	    merge_sides_flat(rights0, rights1);
	  }
	  else {
	    merge_sides(lefts1, lefts);
	    merge_sides(rights0, rights1);
	  }
	  rights0 = rights1;
	}
	layout(ti, *++cit, dtnode, lefts1, rights, scale, flat); // last child
	dtnode_type& dtrightchild = dtnode.children.back();
	int dx = max_overlap(rights0, lefts1, flat);
	dtrightchild.dx = dx+gx;
	dtrightchild.dy = ly+gy;
	translate(dx+gx, 0, lefts1);
	translate(dx+gx, 0, rights);
	if (flat) {
	  merge_sides_flat(lefts1, lefts);
	  merge_sides_flat(rights0, rights);
	}
	else {
	  merge_sides(lefts1, lefts);
	  merge_sides(rights0, rights);
	}  
	int dcx = -dtrightchild.dx/2;
	dtnode.dcx = dcx;
	translate(dcx, ly+gy, lefts);
	translate(dcx, ly+gy, rights);
	lefts.push_front(II(-lx/2,ly));
	rights.push_front(II(lx/2,ly));
      }
    }
  } // draw_tree::layout()

  //! lower_terminals() lowers all terminals so their bottoms are aligned with ymax
  //! (used to draw `flat' trees)
  //
  static void lower_terminals(tree_interface& ti, dtnode_type& dtnode, int y0, int ymax) {
    int y1 = y0 + dtnode.dy;
    if (dtnode.children.empty())
      dtnode.dy += ymax - (y1 + ti.label_height(*dtnode.nodep) + ti.label_descent(*dtnode.nodep));
    else
      for (typename dtnodes_type::iterator it = dtnode.children.begin();
	   it != dtnode.children.end(); ++it)
	lower_terminals(ti, *it, y1, ymax);
  }  // draw_tree::lower_terminals()

  //! draw() does the actual node drawing
  //
  static void draw(tree_interface& ti, const dtnode_type& dtnode, int x0, int y0) {
    int x1 = x0 + dtnode.dx;
    int y1 = y0 + dtnode.dy;
    ti.drawlabel(*dtnode.nodep, x1, y1);
    int x2 = x1 + dtnode.dcx;
    int y2 = y1 + ti.label_height(*dtnode.nodep) + ti.label_descent(*dtnode.nodep);
    for (typename dtnodes_type::const_iterator it = dtnode.children.begin();
	 it != dtnode.children.end(); ++it) {
      ti.drawline(x1, y2, x2+it->dx, y1+it->dy);
      draw(ti, *it, x2, y1);
    }
  } // draw_tree::draw()

public:

  //! draw() lays out and draws a tree on the tree interface ti.
  //! scale rescales the height between nodes.
  //! flat aligns all terminals at the bottom of the diagram.
  //
  static void draw(tree_interface& ti, const node_type& node, 
		   float scale=1.0, bool flat=false) {
    dtnode_type dtnode;
    IIs lefts, rights;
    layout(ti, node, dtnode, lefts, rights, scale, flat);
    assert(!lefts.empty());
    assert(!rights.empty());
    int xmin = util::min_element(lefts, util::first_lessthan())->first;
    int xmax = util::max_element(rights, util::first_lessthan())->first;
    int ymax = lefts.back().second;
    if (flat)
      lower_terminals(ti, dtnode, 0, ymax);
    ti.drawing_area(xmax-xmin+2*ti.border_width(), ymax+2*ti.border_height());
    draw(ti, dtnode.children.front(), ti.border_width()-xmin, ti.border_height());
    ti.flush();
  }  // draw_tree::draw()

}; // draw_tree{}

#endif // DRAW_TREE_H
